# huddle.tcl (working title)
#
# huddle.tcl 0.1.5 2011-08-23 14:46:47 KATO Kanryu(kanryu6@users.sourceforge.net)
#
#   It is published with the terms of tcllib's BSD-style license.
#   See the file named license.terms.
#
# This library provide functions to differentinate string/list/dict in multi-ranks.
#

package require Tcl 8.5
package provide huddle 0.1.7


proc lambda {arguments body args} {
	return [list ::apply [list $arguments $body] {*}$args]
}


namespace eval ::huddle {
	namespace export huddle
	# common subcommands:
	#   get gets strip jsondump set remove
	# type specified subcommands:
	#   create list llength keys
	
	variable mapped_methods
	variable types
}


proc huddle {command args} {
	variable mapped_methods 

	if {[info exists huddle::mapped_methods($command)]} {
		return [tailcall {*}$huddle::mapped_methods($command) {*}$args]
	}
				 		
	return [::huddle::$command {*}$args]

}




proc ::huddle::addType {typeCommand} {
	variable mapped_methods
	variable types
	
	set settings [$typeCommand settings]
	
	if {![dict exists $settings tag]} {
		set $tag [string toupper [dict get $settings type]]
	}
	
	dict with settings {
		foreach subcommand $publicMethods {
			set mapped_methods($subcommand) [lambda args "$typeCommand $subcommand {*}\$args"]
		}

		set types(type:$tag) $type
		set types(callback:$tag) $typeCommand
		set types(isContainer:$tag) $isContainer
		
	}
}

proc ::huddle::isHuddle {arg} {
	if {[lindex $arg 0] ne "HUDDLE" || [llength $arg] != 2} {
		return 0
	}
	variable types
	set subnode [lindex $arg 1]
	set tag [lindex $subnode 0]
	
	if { [array get types "type:$tag"] == ""} {
		return 0
	}
	return 1
}

proc ::huddle::strip {node} {
	variable types
	foreach {head value} $node break
	if {[info exists types(type:$head)]} {
		if $types(isContainer:$head) {
			return [$types(callback:$head) strip_subtags $value]
		} else {
			return $value
		}
	} elseif {$head eq "HUDDLE"} {
		return [strip $value]
	} else {
		error "\{$src\} is not a huddle."
	}

}

proc ::huddle::call {tag cmd arg} {
	variable types
	return [eval $types(callback:$tag) $cmd $arg]
}

proc ::huddle::combine {args} {
	variable types

	foreach {obj} $args {
		checkHuddle $obj
	}

	set first_object [lindex $args 0]
	set unique_tag [lindex [unwrap $first_object] 0]
	
	foreach {obj} $args {
		set node [unwrap  $obj]
		foreach {tag src} $node break

		if {$tag ne $unique_tag} {error "unmatched huddles are given."}
		
		eval lappend result $src
	}
	set src [$types(callback:$tag) append "" {} $result]
	return [wrap [list $tag $src]]
	
}

proc ::huddle::checkHuddle {huddle_object} {
	if {![isHuddle $huddle_object]} {
		error "\{$huddle_object\} is not a huddle."
	}
}

proc ::huddle::to_node {src {tag s}} {

	if {[isHuddle $src]} {
		return [unwrap $src]
	} else {
		return [list $tag $src]
	}
}

proc ::huddle::wrap { node } {
	return [list HUDDLE $node]
}


proc ::huddle::unwrap { huddle_object } {
	return [lindex $huddle_object 1]
}

proc ::huddle::_get_src { huddle_object } {
	return [lindex [unwrap $huddle_object] 1]
}

proc ::huddle::_get {huddle_object args} {

	return [retrieve_huddle $huddle_object $args 0]

}

proc ::huddle::_get_stripped {huddle_object args} {

	return [retrieve_huddle $huddle_object $args 1]
}

proc ::huddle::retrieve_huddle {huddle_object path striped} {
	checkHuddle $huddle_object
	
	set current_node [unwrap $huddle_object]
	
	set target_node [_find_node $current_node $path]
	if $striped {
		return [strip $target_node]
	} else {
		return [wrap $target_node]
	}
	  
}

proc ::huddle::type {huddle_object args} {
	checkHuddle $huddle_object
	variable types
			
	set target_node [_find_node [unwrap $huddle_object] $args]
		
	foreach {tag src} $target_node break
	
	return $types(type:$tag)
}
  
proc ::huddle::_find_node {node path} {
	
	set len [llength $path]
	
	if {$len == 0} {
		return $node	
	} else {
		
		variable types
		
		foreach {tag src} $node break
		
		if {$len == 1} {
			if {![info exists types(type:$tag)]} {error "\{$src\} is not a huddle node."}
	
			return [$types(callback:$tag) get_subnode $src $path]
		} else {
			# length > 1
			set key [lindex $path 0]
			set subpath [lrange $path 1 end]
			if {![info exists types(type:$tag)]} {error "\{$src\} don't have any child node."}
			set subnode [$types(callback:$tag) get_subnode $src $key]
			return [_find_node $subnode $subpath] 
		}
	}
	
}
  


proc ::huddle::equal {obj1 obj2} {
	checkHuddle $obj1
	checkHuddle $obj2
	return [_equal_subnodes [unwrap $obj1] [unwrap $obj2]]
}

proc ::huddle::_equal_subnodes {obj1 obj2} {
	variable types

	foreach {tag1 src1} $obj1 break
	foreach {tag2 src2} $obj2 break
	if {$tag1 ne $tag2} {return 0}
	return [$types(callback:$tag1) equal $src1 $src2]
}

proc ::huddle::_append {objvar args} {
	variable types

	upvar $objvar obj
	checkHuddle $obj
	foreach {tag src} [unwrap $obj] break
	set src [$types(callback:$tag) append $tag $src $args]
	set obj [wrap [list $tag $src]]
	return $obj
}



proc ::huddle::_set {objvar args} {

	upvar $objvar obj
	
	checkHuddle $obj
	set path [lrange $args 0 end-1]

	set new_subnode [to_node [lindex $args end]]

	set node [_apply_to_subnode set [unwrap $obj] [llength $path] $path $new_subnode]
	set obj [wrap $node]
}

proc ::huddle::remove {obj args} {
	checkHuddle $obj
	
	set node [_apply_to_subnode remove [unwrap $obj] [llength $args] $args]
	set obj [wrap $node]
	
	return $obj
}

proc ::huddle::_apply_to_subnode {subcommand node len path {subcommand_arguments ""}} {
	variable types
	foreach {tag src} $node break
	if {$len > 1} {
		set key [lindex $path 0]
		set subpath [lrange $path 1 end]
		incr len -1
		if {![info exists types(type:$tag)]} {error "\{$src\} don't have any child node."}
		set subnode [$types(callback:$tag) get_subnode $src $key]
		set modified_subnode [_apply_to_subnode $subcommand $subnode $len $subpath $subcommand_arguments]
		set src [$types(callback:$tag) set $src $key $modified_subnode]
		return [list $tag $src]
	}
	if {![info exists types(type:$tag)]} {error "\{$src\} is not a huddle node."}

	set src [$types(callback:$tag) $subcommand $src $path $subcommand_arguments]
	return [list $tag $src]
}




proc ::huddle::_dict_type {command args} {
# __TRANSCRIBE_BEGIN__
	switch -- $command {
		settings { ; # type definition
			return {
				type dict
				publicMethods {create keys}
				tag D
				isContainer yes
			}
			# type:   the type-name
			# publicMethods: add methods to huddle's subcommand.
			#		  "get_subnode/strip/set/remove/equal/append" called by huddle module.
			#		  "strip" must be defined at all types.
			#		  "get_subnode" must be defined at isContainer types.
			#		  "set/remove/equal/append" shuould be defined, if you call them.
			# tag:	tag definition("child/parent" word is maybe obsoleted)
		}
		get_subnode { ; # get a sub-node specified by "key" from the tagged-content
			foreach {src key} $args break
			return [dict get $src $key]
		}
		strip_subtags { ; # strip from the tagged-content
			
			set src [lindex $args 0]
			
			foreach {key val} $src {
				lappend result $key [huddle strip $val]
			}
			return $result
		}
		set { ; # set a sub-node from the tagged-content
			foreach {src key value} $args break
			dict set src $key $value
			return $src
		}
		remove { ; # remove a sub-node from the tagged-content
			foreach {src key} $args break
			return [dict remove $src $key]
		}
		equal { ; # check equal for each node
			foreach {src1 src2} $args break
			if {[llength $src1] != [llength $src2]} {return 0}
			foreach {key1 val1} $src1 {
				if {![dict exists $src2 $key1]} {return 0}
				if {![huddle _equal_subnodes $val1 [dict get $src2 $key1]]} {return 0}
			}
			return 1
		}
		append { ; # append nodes
			foreach {str src list} $args break
			if {[llength $list] % 2} {error {wrong # args: should be "huddle append objvar ?key value ...?"}}
			set resultL $src
			foreach {key value} $list {
				if {$str ne ""} {
					lappend resultL $key [huddle to_node $value $str]
				} else {
					lappend resultL $key $value
				}
			}
			return [eval dict create $resultL]
		}
		create { ; # $args: all arguments after "huddle create"
			if {[llength $args] % 2} {error {wrong # args: should be "huddle create ?key value ...?"}}
			set resultL {}
			foreach {key value} $args {
				dict set resultL $key [huddle to_node $value]
			}
			return [huddle wrap [list D $resultL]]
		}
		keys {
			set huddle_object [lindex $args 0]
			
			return [dict keys [_get_src $huddle_object]]
		}
		default {
			error "$command is not callback for dict"
		}
	}
# __TRANSCRIBE_END__
}

proc ::huddle::_list_type {command args} {
	switch -- $command {
		settings {
			return {
				type list
				publicMethods {list llength}
				tag L
				isContainer yes
			}
		}
		get_subnode {
			foreach {src index} $args break
			return [lindex $src $index]
		}
		strip_subtags {
			set src [lindex $args 0]
			set result {}
			foreach {val} $src {
				lappend result [strip $val]
			}
			return $result
		}
		set {
			foreach {src index value} $args break
			lset src $index $value
			return $src
		}
		remove {
			foreach {src index} $args break
			return [lreplace $src $index $index]
		}
		equal {
			foreach {src1 src2} $args break
			if {[llength $src1] != [llength $src2]} {return 0}
			set i 0
			foreach {val1} $src1 {
				if {![huddle _equal_subnodes $val1 [lindex $src2 $i]]} {return 0}
				incr i
			}
			return 1
		}
		append { ; # append nodes
			foreach {str src list} $args break
			set resultL $src
			foreach {value} $list {
				if {$str ne ""} {
					lappend resultL [huddle to_node $value $str]
				} else {
					lappend resultL $value
				}
			}
			return $resultL
		}
		list {
			set resultL {}
			foreach {value} $args {
				lappend resultL [huddle to_node $value]
			}
			return [huddle wrap [list L $resultL]]
		}
		llength {
			set huddle_object [lindex $args 0]
			
			return [llength [_get_src $huddle_object] ]
		}
		default {
			error "$command is not callback for list"
		}
	}
}

proc ::huddle::_string_type {command args} {
	switch -- $command {
		settings {
			return {
				type string
				publicMethods {string}
				tag s
				isContainer no
			}
		}
		
		string {
			return [huddle wrap [list s $args]]
		}
		
		equal {
			foreach {string1 string2} $args break
			return [expr {$string1 eq $string2}]
		}
		
		default {
			error "$command is not callback for string"
		}
	}
}


proc ::huddle::_number_type {command args} {
	switch -- $command {
		settings {
			return {
				type number
				publicMethods {number}
				tag d
				isContainer no
			}
		}
		
		number {
			if {[llength $args] == 1} {
				set argument [lindex $args 0]
				if [string is double -strict $argument] {
					return [huddle wrap [list num $argument]]
				} else {
					error "Argument '$argument' is not a number"
				}
			} else {
				error "Wrong number of arguments for boolean type"
			}
		}
		
		equal {
			foreach {number1 number2} $args break
			return [expr {$number1 == $number2}]
		}
		

		default {
			error "$command is not callback for string"
		}
	}
}


proc ::huddle::_boolean_type {command args} {
	switch -- $command {
		settings {
			return {
				type boolean
				publicMethods {true false}
				tag b
				isContainer no
			}
		}
		
		true {
			if {[llength $args] > 0 } {
				error "boolean type doesn't need arguments"
			}
			
			return [huddle wrap [list b true]]
		}
		
		false {
			if {[llength $args] > 0 } {
				error "boolean type doesn't need arguments"
			}
			
			return [huddle wrap [list b false]]
		}

		
		equal {
			foreach {bool1 bool2} $args break
			return [expr {$bool1 eq $bool2}]
		}
		

		default {
			error "$command is not callback for boolean"
		}
	}
}


proc ::huddle::_null_type {command args} {
	switch -- $command {
		settings {
			return {
				type null
				publicMethods {null}
				tag n
				isContainer no
			}
		}
		
		null {
			if {[llength $args] > 0 } {
				error "null type needs no arguments"
			}
			
			return [huddle wrap [list null]]

		}
		
		equal {
			return 1
		}
		
		default {
			error "$command is not callback for boolean"
		}
	}
}


proc ::huddle::jsondump {huddle_object {offset "  "} {newline "\n"} {begin ""}} {
	variable types
	set nextoff "$begin$offset"
	set nlof "$newline$nextoff"
	set sp " "
	if {[string equal $offset ""]} {set sp ""}
	
	set type [huddle type $huddle_object]
	
	switch -- $type {
		boolean - 
		number -
		null {
			return [huddle strip $huddle_object]
		}
		
		string {
			set data [huddle strip $huddle_object]

			# JSON permits only oneline string
			set data [string map {
					\n \\n
					\t \\t
					\r \\r
					\b \\b
					\f \\f
					\\ \\\\
					\" \\\"
					/ \\/
				} $data
			]
			return "\"$data\""
		}
		list {
			set inner {}
			set len [huddle llength $huddle_object]
			for {set i 0} {$i < $len} {incr i} {
				set subobject [huddle get $huddle_object $i]
				lappend inner [jsondump $subobject $offset $newline $nextoff]
			}
			if {[llength $inner] == 1} {
				return "\[[lindex $inner 0]\]"
			}
			return "\[$nlof[join $inner ,$nlof]$newline$begin\]"
		}
		
		dict {
			set inner {}
			foreach {key} [huddle keys $huddle_object] {
				lappend inner [subst {"$key":$sp[jsondump [huddle get $huddle_object $key] $offset $newline $nextoff]}]
			}
			if {[llength $inner] == 1} {
				return $inner
			}
			return "\{$nlof[join $inner ,$nlof]$newline$begin\}"
		}
		default {
			return [$types(callback:$type) jsondump $data $offset $newline $nextoff]
		}
	}
}

# data is plain old tcl values
# spec is defined as follows:
# {string} - data is simply a string, "quote" it if it's not a number
# {list} - data is a tcl list of strings, convert to JSON arrays
# {list list} - data is a tcl list of lists
# {list dict} - data is a tcl list of dicts
# {dict} - data is a tcl dict of strings
# {dict xx list} - data is a tcl dict where the value of key xx is a tcl list
# {dict * list} - data is a tcl dict of lists
# etc..
proc ::huddle::compile {spec data} {
	while [llength $spec] {
		set type [lindex $spec 0]
		set spec [lrange $spec 1 end]

		switch -- $type {
			dict {
				lappend spec * string

				set result [huddle create]
				foreach {key val} $data {
					foreach {keymatch valtype} $spec {
						if {[string match $keymatch $key]} {
							huddle append result $key [compile $valtype $val]
							break
						}
					}
				}
				return $result
			}
			list {
				if {![llength $spec]} {
					set spec string
				} else {
					set spec [lindex $spec 0]
				}
				set result [huddle list]
				foreach {val} $data {
					huddle append result [compile $spec $val]
				}
				return $result
			}
			string {
#				 if {[string is double -strict $data]} {
#					 return $data
#				 } else {
					return [huddle wrap [list s $data]]
#				 }
			}
			default {error "Invalid type"}
		}
	}
}

namespace eval ::huddle {
	
	array set types {}
	array set callbacks {}

	# Some subcommands conflict reserved words. So, add prefix "_" (e.g. from "set" to "_set")
	array set mapped_methods {
		set			::huddle::_set
		append		::huddle::_append
		get			::huddle::_get
		gets		::huddle::_get_stripped
	}
	
	::huddle::addType ::huddle::_dict_type
	::huddle::addType ::huddle::_list_type
	::huddle::addType ::huddle::_string_type
	::huddle::addType ::huddle::_boolean_type
	::huddle::addType ::huddle::_number_type
	::huddle::addType ::huddle::_null_type

}


